import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Polygon

from dorfperfekt.tile import Terrain

CMAP = LinearSegmentedColormap.from_list("gr", ["g", "w", "r"])

angles = np.linspace(start=0, stop=360, num=6, endpoint=False)
angles = np.deg2rad(angles)

DIST = np.cos(np.deg2rad(30))
XVEC = np.array([0, 1])
YVEC = np.array([DIST, np.sin(np.deg2rad(30))])
VERTICES = np.transpose(np.vstack((np.cos(angles), np.sin(angles))))

TERRAIN_COLORS = {
    Terrain.GRASS: "lightgreen",
    Terrain.FOREST: "darkgreen",
    Terrain.RANCH: "gold",
    Terrain.DWELLING: "firebrick",
    Terrain.WATER: "steelblue",
    Terrain.STATION: "slategray",
    Terrain.TRAIN: "tab:brown",
    Terrain.COAST: "aqua",
}


def pos2coords(pos):
    return np.matmul(pos, np.vstack((XVEC, YVEC))) * 2 * DIST


def coords2pos(coords):
    coords = np.array(coords) / (2 * DIST)
    pos = np.linalg.solve(np.vstack((XVEC, YVEC)).T, coords.T).T
    return tuple([int(np.rint(p)) for p in pos])


def format_map(draw_map):
    def wrapper(*args, **kwargs):
        ax = kwargs["ax"] if "ax" in kwargs else args[0]
        [p.remove() for p in reversed(ax.patches)]
        draw_map(*args, **kwargs)

    return wrapper


@format_map
def draw_position_map(ax, nonruined, ruined, ranked=[], unranked=[]):
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    def draw_tile(pos, color, edge):
        coords = pos2coords(pos)
        inx = (xlim[0] - 1) <= coords[0] <= (xlim[1] + 1)
        iny = (ylim[0] - 1) <= coords[1] <= (ylim[1] + 1)

        if inx and iny:
            corners = coords + VERTICES
            patch = Polygon(corners, edgecolor=edge, facecolor=color)
            ax.add_patch(patch)

    color = CMAP(np.linspace(0, 1, len(ranked)))

    [draw_tile(pos, "lightslategrey", "white") for pos in nonruined]
    [draw_tile(pos, "lightsteelblue", "white") for pos in ruined]

    for idx, pset in enumerate(ranked):
        for pos in pset:
            draw_tile(pos, color[idx], "black")

    [draw_tile(pos, "white", "black") for pos in unranked]


@format_map
def draw_terrain_map(ax, tiles, selected):
    xlim, ylim = ax.get_xlim(), ax.get_ylim()

    for pos, tile in tiles:
        coords = pos2coords(pos)
        inx = (xlim[0] - 1) <= coords[0] <= (xlim[1] + 1)
        iny = (ylim[0] - 1) <= coords[1] <= (ylim[1] + 1)

        if inx and iny:
            corners = coords + VERTICES
            corners = np.vstack((corners, corners[0]))
            corners = np.flipud(corners)
            color = "black" if (pos, tile) == selected else None

            for k in range(len(corners) - 1):
                tcolor = TERRAIN_COLORS[tile.terrains[(k + 2 - tile.ori) % 6]]
                subpatch_corners = [coords, corners[k], corners[k + 1]]
                subpatch = Polygon(subpatch_corners, edgecolor=color, facecolor=tcolor)
                ax.add_patch(subpatch)



from collections import namedtuple

from aenum import OrderedEnum, unique
from cachetools.func import lfu_cache


class InvalidTileDefinitionError(ValueError):
    pass


@unique
class Terrain(OrderedEnum):
    GRASS = "G"
    FOREST = "F"
    RANCH = "R"
    DWELLING = "D"
    WATER = "W"
    STATION = "S"
    TRAIN = "T"
    COAST = "C"
    OPEN = "O"


RESTRICTED_TERRAINS = {Terrain.WATER, Terrain.TRAIN}

RESTRICTED_EXCEPTIONS = [
    {Terrain.WATER, Terrain.STATION},
    {Terrain.WATER, Terrain.COAST},
    {Terrain.TRAIN, Terrain.STATION},
]

PERFECT_ACCEPTIONS = [
    {Terrain.WATER, Terrain.STATION},
    {Terrain.WATER, Terrain.COAST},
    {Terrain.TRAIN, Terrain.STATION},
    {Terrain.COAST, Terrain.GRASS},
    {Terrain.COAST, Terrain.STATION},
    {Terrain.GRASS, Terrain.STATION},
]

Tile = namedtuple("Tile", "terrains ori")


def string2tile(string):
    string = string.upper()
    string = string * 6 if len(string) == 1 else string

    try:
        terrains = tuple(Terrain(t) for t in string)
        return terrains2tile(terrains)
    except ValueError:
        raise InvalidTileDefinitionError


def terrains2tile(terrains):
    if len(terrains) != 6:
        raise InvalidTileDefinitionError

    final_terrains, final_ori = terrains, 0

    for rot_ori in range(1, 6):
        rot_terrains = tuple((*terrains[rot_ori:], *terrains[:rot_ori]))

        if rot_terrains < final_terrains:
            final_terrains, final_ori = rot_terrains, rot_ori

    return Tile(final_terrains, final_ori)


def tile2string(tile):
    string = "".join([t.value for t in tile.terrains])
    return string[-tile.ori :] + string[: -tile.ori]


class SettifyArguments:
    def __init__(self, func):
        self.func = func

    def __call__(self, arg1, arg2):
        return self.func(frozenset((arg1, arg2)))


@SettifyArguments
@lfu_cache(maxsize=64)
def validate_terrains(terrains):
    if Terrain.OPEN in terrains:
        return True, None

    matching = len(terrains) == 1
    restricted = not terrains.isdisjoint(RESTRICTED_TERRAINS)
    excepted = terrains in RESTRICTED_EXCEPTIONS

    accepted = terrains in PERFECT_ACCEPTIONS
    is_invalid = restricted and not (matching or excepted)
    is_perfect = matching or accepted

    return not is_invalid, is_perfect


@SettifyArguments
def validate_tiles(tiles):
    if len(tiles) == 1:
        return True, (True,) * 6

    else:
        is_perfect = [None] * 6
        inner, outer = tuple(tiles)
        for ori in range(6):
            inner_terrain = inner.terrains[(ori - inner.ori) % 6]
            outer_terrain = outer.terrains[(ori - outer.ori) % 6]
            is_valid, is_perfect[ori] = validate_terrains(inner_terrain, outer_terrain)

            if not is_valid:
                return False, None

        return True, tuple(is_perfect)



import sys
import time
from collections import defaultdict
from copy import deepcopy

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from PySide6.QtCore import QObject, QSize, QThread, QTimer, Signal, Slot
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from .display import coords2pos, draw_position_map, draw_terrain_map, pos2coords
from .tile import InvalidTileDefinitionError, string2tile
from .tilemap import InvalidTilePlacementError, TileMap

StyleSheet = """
#BlueProgressBar {
    text-align: center;
    border: 2px solid #2196F3;
    border-radius: 5px;
    background-color: #E0E0E0;
}
#BlueProgressBar::chunk {
    background-color: #2196F3;
    width: 10px; 
    margin: 0.5px;
}
"""


def rescale_axes(fig, ax, origin, scale):
    extent = fig.get_window_extent()
    coords = pos2coords(origin)
    scale = scale * 3 / 4
    ascale = scale * extent.height / extent.width
    ax.set_xlim(coords[0] - scale, coords[0] + scale)
    ax.set_ylim(coords[1] - ascale, coords[1] + ascale)


class Solver(QObject):
    result = Signal(tuple)

    def __init__(self, parent=None, **kwargs):
        super(Solver, self).__init__(parent, **kwargs)

    def interrupt(self):
        self.active = False

    @Slot(tuple)
    def run(self, msg):
        self.active = True
        tilemap, terrains, thresh = msg
        scores = tilemap.scores(terrains, thresh)
        for i, (pos, tilescores) in enumerate(scores):
            self.result.emit((i, pos, tilescores))
            if not self.active:
                break


class MainWindow(QMainWindow):
    run_solver = Signal(tuple)

    def __init__(self):
        QMainWindow.__init__(self)
        self.setMinimumSize(500, 500)

        self.filename = None
        self.tilemap = TileMap()
        self.pos_focus = (0, 0)
        self.ter_focus = (0, 0)

        self.init_menu_bar()
        self.init_window_title()

        # setup top-level layouts
        main_vbox = QVBoxLayout()
        main_hbox = QHBoxLayout()
        side_vbox = QVBoxLayout()
        main_vbox.addLayout(main_hbox)
        main_hbox.addLayout(side_vbox)

        central = QWidget(self)
        central.setLayout(main_vbox)
        self.setCentralWidget(central)

        # setup top-level widgets
        self.init_controls(side_vbox)
        self.terfg, self.terax, self.tercv = self.init_plot_canvas(side_vbox)
        self.tercv.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Expanding)
        self.posfg, self.posax, self.poscv = self.init_plot_canvas(main_hbox)
        self.poscv.callbacks.connect("button_press_event", self.focus)
        self.pgbar = self.init_progress_bar(main_vbox)

        # setup solver thread
        self.thread = QThread()
        self.solver = Solver()
        self.solver.result.connect(self.update_scores)
        self.run_solver.connect(self.solver.run)
        self.solver.moveToThread(self.thread)
        self.thread.start(priority=QThread.Priority.IdlePriority)

        # setup resize event timer
        self.rtimer = QTimer()
        self.rtimer.setSingleShot(True)
        self.rtimer.timeout.connect(self.resized)

        # setup progress timer
        self.ptimer = QTimer()
        self.ptimer.timeout.connect(self.update_progress)

    def init_window_title(self):
        filestring = " (" + self.filename + ")" if self.filename is not None else ""
        self.setWindowTitle("Dorfperfekt" + filestring + "[*]")

    def init_menu_bar(self):
        menu = self.menuBar()
        file_menu = menu.addMenu("File")

        open_action = file_menu.addAction("Open...")
        open_action.triggered.connect(self.open)

        save_action = file_menu.addAction("Save")
        save_action.setShortcut("Ctrl+S")
        save_action.triggered.connect(self.save)

        saveas_action = file_menu.addAction("Save as...")
        saveas_action.triggered.connect(self.saveas)

    def init_plot_canvas(self, parent):
        fig = Figure()
        ax = fig.add_axes([0, 0, 1, 1])
        ax.axis("off")
        canvas = FigureCanvas(fig)
        parent.addWidget(canvas, stretch=1)

        return fig, ax, canvas

    def init_controls(self, parent):
        grid = QGridLayout()

        grid.addWidget(QLabel("Position Map Horizontal Scale :"), 0, 0, 1, 2)
        grid.addWidget(possc := QSpinBox(minimum=10), 0, 2)
        grid.addWidget(QLabel("Terrain Map Horizontal Scale :"), 1, 0, 1, 2)
        grid.addWidget(tersc := QSpinBox(minimum=5), 1, 2)

        grid.addWidget(set_origin := QPushButton("Set Origin"), 2, 0)
        grid.addWidget(rst_origin := QPushButton("Reset Origin"), 2, 1)
        grid.addWidget(refresh := QPushButton("Refresh"), 2, 2)

        grid.addWidget(QLabel("Tile Definition String :"), 3, 0, 1, 2)
        grid.addWidget(ledit := QLineEdit(), 3, 2)
        ledit.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Preferred)
        grid.addWidget(QLabel("Tile Counter Threshold :"), 4, 0, 1, 2)
        grid.addWidget(thresh := QSpinBox(minimum=1), 4, 2)

        grid.addWidget(solve := QPushButton("Solve"), 5, 0, 1, 3)
        grid.addWidget(delete := QPushButton("Delete"), 6, 0)
        grid.addWidget(rotate := QPushButton("Rotate"), 6, 1)
        grid.addWidget(place := QPushButton("Place"), 6, 2)

        grid.addWidget(total := QLabel(), 7, 0, 1, 3)
        grid.addWidget(cover := QLabel(), 8, 0, 1, 3)

        possc.valueChanged.connect(self.draw_position_map)
        tersc.valueChanged.connect(self.draw_terrain_map)
        refresh.clicked.connect(self.draw_position_map)
        set_origin.clicked.connect(lambda: self.change_origin(self.ter_focus))
        rst_origin.clicked.connect(lambda: self.change_origin((0, 0)))

        solve.clicked.connect(self.solve)
        place.clicked.connect(self.place)
        rotate.clicked.connect(self.rotate)
        delete.clicked.connect(self.delete)

        parent.addLayout(grid)

        self.possc = possc
        self.tersc = tersc
        self.ledit = ledit
        self.thresh = thresh
        self.total = total
        self.cover = cover

    def init_progress_bar(self, parent):
        pgbar = QProgressBar(self, objectName="BlueProgressBar")
        parent.addWidget(pgbar)
        return pgbar

    def closeEvent(self, event):
        if self.isWindowModified():
            ret = QMessageBox.warning(
                self,
                "Dorfperfekt -- Warning",
                "Current file is not saved!",
                QMessageBox.Ok | QMessageBox.Cancel,
            )
            if ret is QMessageBox.Cancel:
                event.ignore()
                return

        self.solver.interrupt()
        self.thread.quit()
        self.thread.wait()
        event.accept()

    def resizeEvent(self, event):
        self.rtimer.start(500)
        event.accept()

    def resized(self):
        self.draw_position_map()
        self.draw_terrain_map()

    def reset(self, modified=False):
        self.solver.interrupt()
        self.ptimer.stop()
        self.scores = dict()
        self.tile2solve = None
        self.tile2place = None
        self.resized()
        self.setWindowModified(modified)
        self.ledit.setText("")
        self.pgbar.setValue(0)

        fstring = "{placed:d} placed. {ruined:d} ruined. {unique:d} unique."
        self.total.setText(
            fstring.format(
                placed=len(self.tilemap),
                ruined=len(self.tilemap.ruined),
                unique=len(self.tilemap.counter),
            )
        )

        self.cover.setText("")

    def change_origin(self, origin):
        self.pos_focus = origin
        self.ter_focus = origin
        self.tile2place = None
        self.resized()

    def draw_position_map(self):
        ruined = set(self.tilemap.ruined)
        nonruined = set(self.tilemap) - ruined

        ranked = defaultdict(set)
        unranked = set()
        for pos, tilescores in self.scores.items():
            if tilescores is None:
                unranked.add(pos)
            else:
                score = min(score for score, _ in tilescores)
                ranked[score].add(pos)

        ranked = [ranked[score] for score in sorted(ranked)]

        rescale_axes(
            fig=self.posfg,
            ax=self.posax,
            origin=self.pos_focus,
            scale=self.possc.value(),
        )

        draw_position_map(self.posax, nonruined, ruined, ranked, unranked)

        self.poscv.draw()
        self.poscv.flush_events()

    def draw_terrain_map(self):
        tiles = list(self.tilemap.items())
        if self.tile2place is not None:
            selected = (self.ter_focus, self.tile2place)
            tiles.append(selected)
        else:
            selected = None

        rescale_axes(
            fig=self.terfg,
            ax=self.terax,
            origin=self.ter_focus,
            scale=self.tersc.value(),
        )

        draw_terrain_map(self.terax, tiles, selected)

        self.tercv.draw()
        self.tercv.flush_events()

    def focus(self, event):
        if event.inaxes is not None:
            self.ter_focus = coords2pos((event.xdata, event.ydata))

            if self.ter_focus in self.tilemap:
                # focus position is an existing tile
                self.tile2place = self.tilemap[self.ter_focus]

            elif self.ter_focus in self.scores:
                # focus position is an open position
                tilescores = self.scores[self.ter_focus]

                if tilescores is None:
                    # focus position has not been scored, use default
                    self.tile2place = self.tile2solve
                else:
                    # focus position has been scored, get best score
                    scores, tiles = zip(*tilescores)
                    idx = scores.index(min(scores))
                    self.tile2place = tiles[idx]

            else:
                # focus position is not a valid tile
                self.tile2place = None

            self.draw_terrain_map()

    def open(self):
        if self.isWindowModified():
            ret = QMessageBox.warning(
                self,
                "Dorfperfekt -- Warning",
                "Current file is not saved!",
                QMessageBox.Ok | QMessageBox.Cancel,
            )
            if ret is QMessageBox.Cancel:
                return

        filename = QFileDialog.getOpenFileName(self)[0]
        if filename:
            self.tilemap = TileMap.from_file(filename)
            self.filename = filename
            self.init_window_title()
            self.reset()

    def save(self):
        if not self.isWindowModified():
            return
        elif self.filename is None:
            self.saveas()
        else:
            self.tilemap.write_file(self.filename)
            self.setWindowModified(False)

    def saveas(self):
        filename = QFileDialog.getSaveFileName(self)[0]
        if filename:
            self.tilemap.write_file(filename)
            self.filename = filename
            self.init_window_title()
            self.setWindowModified(False)

    def solve(self):
        try:
            string = self.ledit.text()
            self.tile2solve = string2tile(string)
        except InvalidTileDefinitionError:
            return

        cov = len(
            [
                tile
                for tile, count in self.tilemap.counter.items()
                if count >= self.thresh.value()
            ]
        )
        self.cover.setText("Solver considering {:d} unique tiles.".format(cov))

        self.pgbar.setMaximum(len(self.tilemap.open))
        self.scores = {pos: None for pos in self.tilemap.open}
        self.solver.interrupt()
        self.run_solver.emit(
            (self.tilemap, self.tile2solve.terrains, self.thresh.value())
        )
        self.progress = 0
        self.ptimer.start(2000)

    def update_scores(self, msg):
        i, pos, tilescores = msg
        if tilescores:
            self.scores[pos] = tilescores
        else:
            del self.scores[pos]

        self.progress = i + 1

    def update_progress(self):
        self.pgbar.setValue(self.progress)
        if self.progress == self.pgbar.maximum():
            self.ptimer.stop()

    def place(self):
        if self.tile2place is not None and self.ter_focus not in self.tilemap:
            try:
                self.tilemap[self.ter_focus] = self.tile2place
                self.reset(modified=True)
            except InvalidTilePlacementError:
                pass

    def rotate(self):
        if self.tile2place is not None and self.ter_focus not in self.tilemap:
            newori = (self.tile2place.ori + 1) % 6
            self.tile2place = self.tile2place._replace(ori=newori)
            self.draw_terrain_map()

    def delete(self):
        if self.tile2place is not None and self.ter_focus in self.tilemap:
            del self.tilemap[self.ter_focus]
            self.reset(modified=True)


def main():
    app = QApplication(sys.argv)
    app.setStyleSheet(StyleSheet)
    main_window = MainWindow()
    main_window.show()
    main_window.reset()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dorfperfekt']

package_data = \
{'': ['*']}

install_requires = \
['PySide6>=6.2.3,<7.0.0',
 'aenum>=3.1.8,<4.0.0',
 'cachetools>=5.0.0,<6.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.2,<2.0.0']

entry_points = \
{'console_scripts': ['dorfperfekt = dorfperfekt.__main__:main']}

setup_kwargs = {
    'name': 'dorfperfekt',
    'version': '0.2.2',
    'description': 'Tile placement suggestions for the game Dorfromantik.',
    'long_description': '# Dorfperfekt\n\nTile placement suggestions for the game Dorfromantik. With an emphasis on perfect tile placement, this software tool enables the player to play indefinitely (and achieve a massive score while doing so).\n\n![demo dorfperfekt](https://github.com/amosborne/dorfperfekt/raw/main/demo_dorfperfekt.png)\n![demo dorfromantik map](https://github.com/amosborne/dorfperfekt/raw/main/demo_dorfromantik_map.png)\n![demo dorfromantik score](https://github.com/amosborne/dorfperfekt/raw/main/demo_dorfromantik_score.png)\n\n## How It Works\n\nThe player inputs a six-character text string representing the next tile to be placed. Each valid placement is then evaluated by the solver according to the following (in order of precedence):\n\n1. How many tiles would be ruined by this placement? Less is better.\n2. Of the tiles encountered so far, how many would alternatively fit perfectly at this position? Less is better.\n3. Assuming this placement is made, consider each adjacent open position. Of the tiles encountered so far, how many would fit perfectly at that adjacent position? Take the minimum of all adjacent open positions. More is better.\n\nThe latter two computations will take a significant amount of time as the game progresses and more unique tiles are encountered. A threshold is set by the user so that the solver can be instructed to ignore rarer tiles.\n\n## Using the Application\n\nDorfperfekt displays an overall map of the board. Dark gray tiles are non-ruined placements and light gray tiles are ruined placements. After entering a tile definition string and pressing the solve button, the progress bar will increment and pressing the refresh button will overlay a heatmap of the possible moves onto the board. Green is better, red is worse, and white is neutral (or not yet evaluated).\n\nPositions on the map can be clicked on. By clicking on a proposed position for your next placement, a view of the local terrain is generated. The tile to be placed is given a proposed best rotation but the user may use the rotate button to select an alternate rotation.\n\nExisting tiles may also be clicked on such that they may be deleted or set as the origin to recenter the map of the board.\n\n## Tile Definitions\n\nA tile is defined by a six-character text string where each character represents the edge terrains in **clockwise** order. If all edges of the tile are the same a single character may be used instead. Tile characters are deliberately selected to all be accessible from the left hand.\n\n- Grass, "g"\n- Forest, "f"\n- Ranch, "r" (ie. wheat/lavender fields)\n- Dwelling, "d" (ie. houses)\n- Water, "w" (ie. rivers)\n- Station, "s"\n- Train, "t"\n- Coast, "c" (ie. lakes)\n\n## Installation\n\nFrom the command line, install with `pip install dorfperfekt`. The application can then be run from the command line with `dorfperfekt`.\n\n## Development\n\nSetting up the software development environment is easy.\n\n```bash\npoetry install\npoetry run pre-commit install\n```\n',
    'author': 'amosborne',
    'author_email': 'amosborne@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/amosborne/dorfperfekt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



import re
from collections import Counter, OrderedDict
from collections.abc import MutableMapping
from math import inf
from multiprocessing import Pool, cpu_count

from cachetools.func import lru_cache

from .tile import Terrain, Tile, string2tile, terrains2tile, tile2string, validate_tiles


class InvalidTilePlacementError(ValueError):
    pass


OFFSETS = [(1, 0), (0, 1), (-1, 1), (-1, 0), (0, -1), (1, -1)]

# FUTURE IDEAS:
#   1. self.ruined = set()
#   3. self.solve(terrains, thresh1=1, thresh2=1)


@lru_cache(maxsize=64)
def adjacent_positions(pos):
    return [(pos[0] + off[0], pos[1] + off[1]) for off in OFFSETS]


class TileMap(MutableMapping):
    def __init__(self):
        self.tiles = OrderedDict()
        self.counter = Counter()
        self.ruined = list()
        self.open = set([(0, 0)])
        self[0, 0] = string2tile("g")

    @staticmethod
    def from_file(filepath):
        tilemap = TileMap()
        del tilemap[0, 0]
        pattern = r"^([GFRDWSTC]{6}) (-?\d+) (-?\d+)$"
        with open(filepath) as file:
            for line in file:
                match = re.match(pattern, line)
                tile = string2tile(match[1])
                pos = (int(match[2]), int(match[3]))
                tilemap[pos] = tile

        return tilemap

    def write_file(self, filepath):
        with open(filepath, "w") as file:
            fstring = "{} {:d} {:d}\n"
            for pos, tile in self.tiles.items():
                line = fstring.format(tile2string(tile), *pos)
                file.write(line)

    def __setitem__(self, pos, tile):
        outer = self.outer_tile(pos)
        valid, perfect = validate_tiles(tile, outer)

        if not (valid and pos in self.open):
            raise InvalidTilePlacementError

        self.tiles[pos] = tile
        self.open.remove(pos)
        self.counter[tile.terrains] += 1

        adj = adjacent_positions(pos)

        for adj_pos, adj_perfect in zip(adj, perfect):
            if adj_perfect is False:
                self.ruined.extend([pos, adj_pos])

            if adj_pos not in self:
                self.open.add(adj_pos)

    def __getitem__(self, pos):
        return self.tiles[pos]

    def __delitem__(self, pos):
        inner = self[pos]
        outer = self.outer_tile(pos)
        _, perfect = validate_tiles(inner, outer)

        del self.tiles[pos]
        self.open.add(pos)

        count = self.counter[inner.terrains]
        if count == 1:
            del self.counter[inner.terrains]
        else:
            self.counter[inner.terrains] = count - 1

        adj = adjacent_positions(pos)

        for adj_pos, adj_perfect in zip(adj, perfect):
            if adj_perfect is False:
                self.ruined.remove(pos)
                self.ruined.remove(adj_pos)

            if adj_pos not in self:
                adj_adj = adjacent_positions(adj_pos)
                found = any([pos in self for pos in adj_adj])
                if not found:
                    self.open.discard(adj_pos)

    def __iter__(self):
        return self.tiles.__iter__()

    def __len__(self):
        return len(self.tiles)

    def outer_tile(self, pos):
        terrains = [Terrain.OPEN] * 6
        for ori, adj in enumerate(adjacent_positions(pos)):
            if adj in self:
                tile = self[adj]
                terrains[ori] = tile.terrains[(ori - tile.ori + 3) % 6]

        return terrains2tile(tuple(terrains))

    def perfect_alternates(self, pos, thresh=1):
        count = 0
        pre_ruined = len(set(self.ruined))
        for terrains, subcount in self.counter.items():
            if subcount < thresh:
                continue

            for ori in range(6):
                try:
                    self[pos] = Tile(terrains, ori)
                    post_ruined = len(set(self.ruined))
                    del self[pos]
                    newly_ruined = post_ruined - pre_ruined
                    if not newly_ruined:
                        count += subcount
                        break
                except InvalidTilePlacementError:
                    pass

        return count

    def score_tile(self, pos, tile, thresh=1):
        pre_ruined = len(set(self.ruined))

        self[pos] = tile

        open_adj = [adj for adj in adjacent_positions(pos) if adj in self.open]
        secondorder_alternates = (
            min([self.perfect_alternates(adj, thresh) for adj in open_adj])
            if open_adj
            else inf
        )

        post_ruined = len(set(self.ruined))

        del self[pos]

        newly_ruined = post_ruined - pre_ruined

        alternates = self.perfect_alternates(pos, thresh)

        return newly_ruined, alternates, -secondorder_alternates

    def score_pos(self, args):
        pos, terrains, thresh = args
        scores = set()
        for ori in range(6):
            try:
                tile = Tile(terrains, ori)
                score = self.score_tile(pos, tile, thresh)
                scores.add((score, tile))
            except InvalidTilePlacementError:
                pass

        return pos, scores

    def scores(self, terrains, thresh=1):
        args = ((pos, terrains, thresh) for pos in self.open)
        with Pool(cpu_count() // 2) as pool:
            for score in pool.imap_unordered(self.score_pos, args):
                yield score
